from typing import Any, Callable

__version__: str
Schema: Any
Table: Any
RecordBatch: Any
RecordBatchReader: Any
Field: Any
DataType: Any
ListType: Any
StructType: Any
MapType: Any
schema: Any
map_: Any
list_: Any
field: Any
struct: Any
type_for_alias: Any
date32: Any
date64: Any
decimal128: Any
int32: Any
float16: Any
float32: Any
float64: Any
dictionary: Any
timestamp: Any
TimestampType: Any

py_buffer: Callable[[bytes], Any]
NativeFile: Any
BufferReader: Any
PythonFile: Any
